/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.agent;

import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveEntity;
import org.xmloperator.reactive.exception.AgentAlreadyLaunchedException;

/**
 * Agent implementation.
 */
public abstract class AgentImpl implements Agent {

	protected String agentName = this.getClass().getName();
	protected LaunchingContext launchingContext = null;
	protected ReactiveEntity entity = null;
	protected PrintStream trace = null;
	
	public AgentImpl() {
	}
	
	/**
	 * Add a long value parameter to the name of this Agent.
	 * This is optional but must be done within constructor.
	 * @param value a long value.
	 */
	protected final void addParameter(long value) {
		if (this.entity != null)
			throw new IllegalStateException();
		StringBuffer sb = new StringBuffer(this.agentName);
		sb.append('/');
		sb.append(value);
		this.agentName = sb.toString();
	}
	
	/**
	 * Add a string value parameter to the name of this Agent.
	 * This is optional but must be done within constructor.
	 * @param value a string value.
	 */
	protected final void addParameter(String value) {
		if (value.indexOf('"') > 0)
			throw new IllegalArgumentException();
		if (this.entity != null)
			throw new IllegalStateException();
		StringBuffer sb = new StringBuffer(this.agentName);
		sb.append('/');
		sb.append('"');
		sb.append(value);
		sb.append('"');
		this.agentName = sb.toString();
	}
	
	/**
	 * Returns the ReactiveEntity associated with this Agent.
	 * Access to this method is limited to the package of its class.
	 * @return the ReactiveEntity associated with this Agent.
	 */
	final ReactiveEntity getReactiveEntity() {
		return this.entity;
	}
	
	/**
	 * Returns the status of this agent.
	 * @return the status of this agent.
	 */
	public final int getStatus() {
		return this.launchingContext != null ? this.launchingContext
				.getAgentStatus() : LaunchingContext.STATUS_BEING_LAUNCHED;
	}

	///////////////////////
	// Agent implementation
	///////////////////////
	
	/**
	 * Connects to the ReactiveCommunity given by the LaunchingContext.
	 */	
	public void init(LaunchingContext launchingContext) {
		if (launchingContext == null)
			throw new IllegalArgumentException();
		if (this.launchingContext != null || this.entity != null)
			throw new AgentAlreadyLaunchedException();
		this.launchingContext = launchingContext;
		this.entity = launchingContext.getReactiveEntity();
		this.trace = launchingContext.getTrace();
		ReactiveCommunity community = launchingContext.getReactiveCommunity();
		this.entity.connect(community);
	}

	/**
	 * Disconnects the entity from the community.
	 */
	public void terminate() {
		this.entity.disconnect();
	}
	
	////////////////////////
	// Object implementation
	////////////////////////

	public final String toString() {
		return this.entity != null ? this.entity.getEntityId() + "("
				+ this.agentName + ")" : this.agentName;
	}
}
