/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.ball;

import java.util.List;

import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.MessageEvent;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.SimpleMessage;

import org.xmloperator.reactive.agent.AgentImpl;

import org.xmloperator.reactive.impl.SimpleMessageImpl;

/**
 * This Agent plays to pass each received ball to the next player.
 * It aims to test Message processing.
 */
public final class BallPlayerAgent extends AgentImpl {
	private final int playerCount;
	private final int playerNumber;
	private final int stepCount;
	private final long startInstant;
	private final MessageEvent declarationMessageEvent;
	private final MessageEvent passBallMessageEvent;
	private final String[] playerEntityIds;
	private String nextPlayerEntityId;
	private String previousPlayerEntityId;
	
	public BallPlayerAgent(int playerCount, int playerNumber, int stepCount,
			long startInstant, MessageEvent declarationMessageEvent,
			MessageEvent passBallMessageEvent) {
		if (playerCount < 2 || playerNumber < 0 || stepCount < 0
				|| startInstant < 1)
			throw new IllegalArgumentException();
		this.playerCount = playerCount;
		this.playerNumber = playerNumber;
		this.stepCount = stepCount;
		this.startInstant = startInstant;
		this.declarationMessageEvent = declarationMessageEvent;
		this.passBallMessageEvent = passBallMessageEvent;
		this.addParameter(playerNumber);
		this.playerEntityIds = new String[playerCount];
	}

	private final int playerNumber(String playerEntityId) {
		for (int k = 0; k < this.playerCount; k++)
			if (this.playerEntityIds[k] == playerEntityId)
				return k;
		throw new IllegalArgumentException();
	}
	
	private final String nextPlayerEntityId(String playerEntityId) {
		int playerNumber = this.playerNumber(playerEntityId);
		playerNumber = (playerNumber + 1) % this.playerCount;
		return this.playerEntityIds[playerNumber];
	}

	private final String previousPlayerEntityId(String playerEntityId) {
		int playerNumber = this.playerNumber(playerEntityId);
		playerNumber = (this.playerCount + playerNumber - 1) % this.playerCount;
		return this.playerEntityIds[playerNumber];
	}

	private final void passBallTo(String destinationEntityId) {
		this.entity.generateMessageEvent(this.passBallMessageEvent,
				new SimpleMessageImpl<String>(destinationEntityId));
		long instant = this.entity.getCommunity().getInstant();
		if (this.trace != null)
			this.trace.println(instant + ": " + this.entity + ">" + destinationEntityId + " ");
	}
	
	///////////////////////
	// Agent implementation
	///////////////////////
	
	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		ReactiveCommunity community = launchingContext.getReactiveCommunity();
		while (community.getInstant() < this.startInstant)
			this.entity.stop();
		/*
		 * Declares associations between playerNumbers and entityIds.
		 */
		this.entity.generateMessageEvent(this.declarationMessageEvent,
				new SimpleMessageImpl<Integer>(
						new Integer(this.playerNumber)));
		this.entity.stop();
		List<Event> eventList = this.entity.suspend();
		if (eventList.size() != 1)
			throw new IllegalStateException();
		MessageEvent messageEvent = (MessageEvent) eventList.get(0);
		if (messageEvent != this.declarationMessageEvent)
			throw new IllegalStateException();
		List<Message> messageList = messageEvent.getMessageList();
		if (messageList.size() != this.playerCount)
			throw new IllegalStateException();
		for (Message message : messageList) {
			SimpleMessage<?> playerMessage =
				(SimpleMessage<?>) message;
			String entityId = playerMessage.getSenderEntityId();
			int player = ((Integer)playerMessage.getObject()).intValue();
			this.playerEntityIds[player] = entityId;
		}
		String entityId = this.entity.getEntityId();
		this.nextPlayerEntityId = nextPlayerEntityId(entityId);
		this.previousPlayerEntityId = previousPlayerEntityId(entityId);
		/*
		 * The player 0 starts by passing a ball forward and another ball backward.
		 */
		if (this.playerNumber == 0) {
			this.passBallTo(this.nextPlayerEntityId);
			this.passBallTo(this.previousPlayerEntityId);
		}
		this.entity.stop();
	}

	public void participate() {
		for (int step = 0; step < this.stepCount; step++) {
			List<Event> eventList = this.entity.suspend();
			for (Event event : eventList)
				if (event.isMessageEvent()) {
					MessageEvent messageEvent = (MessageEvent) event;
					if (messageEvent != this.passBallMessageEvent)
						throw new IllegalStateException();
					for (Message message : messageEvent.getMessageList()) {
						SimpleMessage<?> playerMessage = (SimpleMessage<?>) message;
						String senderEntityId = playerMessage.getSenderEntityId();
						String destinationEntityId = ((String)playerMessage.getObject());
						if (destinationEntityId == this.entity.getEntityId()) {
							/*
							 * Resends the ball to the new destination.
							 */
							this.passBallTo(senderEntityId == this.previousPlayerEntityId
								? this.nextPlayerEntityId
								: this.previousPlayerEntityId);
						}
					}
				}
			this.entity.stop();
		}
	}
}
