/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.action;

import java.math.BigInteger;

import org.xmloperator.reactive.demo.market.Account;
import org.xmloperator.reactive.demo.market.Anticipation;
import org.xmloperator.reactive.demo.market.Proposal;
import org.xmloperator.reactive.demo.market.Resource;

/**
 * Action of ordering some Resource accordingly to a Proposal. Order is given
 * to transfer the amount.
 */
public final class OrderAction extends OperatorAction {
	private static int orderCount = 0;
	
	private final class NegativeAnticipation implements Anticipation {

		//////////////////////////////
		// Anticipation implementation
		//////////////////////////////
		
		public Resource getResource() {
			return proposal.getPaymentResource();
		}

		public BigInteger getResourceUnitCount() {
			return orderFactor.multiply(BigInteger
					.valueOf(-proposal.getPaymentResourceAtomicUnitCount()));
		}
		
		////////////////////////
		// Object implementation
		////////////////////////
		
		public String toString() {
			return "Order" + orderId + " NegativeAnticipation"
					+ "(" + this.getResource() + ")";
		}
	}

	private final class PositiveAnticipation implements Anticipation {

		//////////////////////////////
		// Anticipation implementation
		//////////////////////////////
		
		public Resource getResource() {
			return proposal.getProposedResource();
		}

		public BigInteger getResourceUnitCount() {
			return orderFactor.multiply(BigInteger
					.valueOf(proposal.getProposedResourceAtomicUnitCount()));
		}
		
		////////////////////////
		// Object implementation
		////////////////////////
		
		public String toString() {
			return "Order" + orderId + " PositiveAnticipation"
					+ "(" + this.getResource() + ")";
		}
	}

	private final int orderId = ++orderCount;
	
	public final BigInteger orderFactor;
	public final Account paymentAccount;
	public final Account deliveryAccount;
	
	private final Anticipation negativeAnticipation = new NegativeAnticipation();
	private final Anticipation positiveAnticipation = new PositiveAnticipation();
	
	public OrderAction(long operatorKey, Proposal proposal,
			BigInteger factor, Account paymentAccount, Account deliveryAccount) {
		super(operatorKey, proposal);
		if (factor.compareTo(BigInteger.ONE) < 0
				|| paymentAccount == null || deliveryAccount == null)
			throw new IllegalArgumentException();
		if (paymentAccount.getResource() != proposal.getPaymentResource()
				|| deliveryAccount.getResource() != proposal
						.getProposedResource())
			throw new IllegalArgumentException();
		this.orderFactor = factor;
		this.paymentAccount = paymentAccount;
		this.deliveryAccount = deliveryAccount;
	}
	
	public final Anticipation getNegativeAnticipation() {
		return this.negativeAnticipation;
	}
	
	public final Anticipation newPositiveAnticipation() {
		return this.positiveAnticipation;
	}
	
	////////////////////////
	// Action implementation
	////////////////////////
	
	public void execute(long permissionKey, long instant) {
		super.execute(permissionKey, instant);
		BigInteger unitCount = this.orderFactor.multiply(
				BigInteger.valueOf(this.proposal.getPaymentResourceAtomicUnitCount()));
		this.proposal.getMarketBoard().newTransfer(permissionKey,
				this.operatorKey,
				this.paymentAccount, this.proposal.getPaymentDestinationAccount(),
				unitCount).execute();
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		return "Order(" + this.proposal + "," + this.orderFactor + ")";
	}
}
