/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.impl;

import java.math.BigInteger;
import java.util.concurrent.atomic.AtomicInteger;

import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.demo.market.Account;
import org.xmloperator.reactive.demo.market.Bank;
import org.xmloperator.reactive.demo.market.DifferentResourcesException;
import org.xmloperator.reactive.demo.market.IncorrectAccountHolderException;
import org.xmloperator.reactive.demo.market.Resource;
import org.xmloperator.reactive.exception.IncorrectKeyException;

class AccountImpl implements Account {
	private static AtomicInteger accountCount = new AtomicInteger();
	protected static PrintStream trace = null;

	static void setTrace(PrintStream trace) {
		AccountImpl.trace = trace;
	}

	protected final int id = accountCount.incrementAndGet();
	protected final Bank bank;
	protected final long holderKey;
	protected BigInteger resourceUnitCount = BigInteger.ZERO;
	
	/**
	 * Constructor.
	 * @param bank the Bank that manages this Account.
	 * @param holderEntity the ReactiveEntity of the holder of this Account.
	 */
	AccountImpl(Bank bank, long holderKey) {
		this.bank = bank;
		this.holderKey = holderKey;
	}

	/**
	 * Adds a count of Resource units to this Account.
	 * @param unitCount a positive count of Resource units.
	 */
	protected final void add(BigInteger unitCount) {
		synchronized (this) {
			this.resourceUnitCount = this.resourceUnitCount.add(unitCount);
		}
	}
	
	/**
	 * Subtracts a count of Resource units from this Account.
	 * @param unitCount a count of Resource units.
	 * @return true in case of success, false on failure due to insufficient sold.
	 */
	protected boolean subtract(BigInteger unitCount)
			throws IncorrectAccountHolderException {
		synchronized (this) {
			BigInteger resourceUnitCount = this.resourceUnitCount.add(unitCount
					.negate());
			if (resourceUnitCount.compareTo(BigInteger.ZERO) < 0)
				return false;
			this.resourceUnitCount = resourceUnitCount;
		}
		return true;
	}

	////////////////////////
	// Account implementation
	/////////////////////////

	public final Resource getResource() {
		return this.bank.getResource();
	}
	
	public final Bank getBank() {
		return this.bank;
	}

	public BigInteger getMaxDebt() {
		return BigInteger.ZERO;
	}

	public final BigInteger getResourceUnitCount(long holderKey) {
		if (holderKey != this.holderKey)
			throw new IncorrectAccountHolderException();
		return this.resourceUnitCount;
	}

	public final double getResourceValue(long holderKey)
			throws IncorrectKeyException {
		return this.resourceUnitCount.doubleValue()
				* this.getResource().getUnitValue();
	}
	
	public final boolean transfer(long holderKey,
			Account destinationAccount, BigInteger unitCount)
			throws IncorrectAccountHolderException, DifferentResourcesException {
		if (destinationAccount == null
				|| unitCount.compareTo(BigInteger.ZERO) < 0)
			throw new IllegalArgumentException();
		if (destinationAccount.getResource() != this.getResource())
			throw new DifferentResourcesException();
		if (holderKey != this.holderKey)
			throw new IncorrectAccountHolderException();
		if (!this.subtract(unitCount))
			return false;
		((AccountImpl)destinationAccount).add(unitCount);
		return true;
	}
	
	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		return "Account" + this.id + "(" + this.getResource() + ")";
	}
}
