/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.demo.market.impl;

import java.math.BigInteger;
import java.util.Random;

import org.xmloperator.reactive.demo.market.Ingredient;
import org.xmloperator.reactive.demo.market.ProjectedAccount;
import org.xmloperator.reactive.demo.market.Resource;

/**
 * Ingredient implementation.
 */
public class IngredientImpl implements Ingredient {
	protected final Resource resource;
	protected int resourceUnitCount;
	protected final ProjectedAccount affectedAccount;

	/**
	 * Constructor.
	 * 
	 * @param resource
	 *            a Resource.
	 * @param resourceUnitCount
	 *            a number of Resource units. The number is positive for a
	 *            produced Resource, negative for a consumed Resource.
	 * @param affectedAccount
	 *            an ProjectedAccount for the Resource.
	 */
	public IngredientImpl(Resource resource, int resourceUnitCount,
			ProjectedAccount affectedAccount) {
		if (resource == null || affectedAccount == null
				|| affectedAccount.getResource() != resource)
			throw new IllegalArgumentException();
		this.resource = resource;
		this.resourceUnitCount = resourceUnitCount;
		this.affectedAccount = affectedAccount;
	}

	/**
	 * Applies this Ingredient multiplied by the given factor on the given Account.
	 * @param factor a multiplicative factor.
	 * @return true in case of success, false in case of insufficient sold.
	 */
	final boolean affectAccount(BigInteger factor) {
		if (factor.compareTo(BigInteger.ONE) < 0)
			throw new IllegalArgumentException();
		BigInteger unitCount = factor.multiply(BigInteger
				.valueOf(Math.abs(this.resourceUnitCount)));
		if (this.resourceUnitCount < 0)
			return ((AccountImpl)this.affectedAccount).subtract(unitCount);
		((AccountImpl)this.affectedAccount).add(unitCount);
		return true;
	}
	
	/**
	 * Adjusts randomly the Resource unit count of this Ingredient.
	 * @param random an random generator.
	 */
	final void adjustResourceUnitCount(Random random) {
		int increment = this.resourceUnitCount / 51;
		if (increment != 0)
			this.resourceUnitCount += (random.nextInt(3) - 1) * increment;
	}

	////////////////////////////
	// Ingredient implementation
	////////////////////////////
	
	public final Resource getResource() {
		return this.resource;
	}

	public final int getResourceUnitCount() {
		return this.resourceUnitCount;
	}

	public final double getIngredientValue() {
		return this.resourceUnitCount * this.resource.getUnitValue();
	}
	
	public final ProjectedAccount getAffectedAccount() {
		return this.affectedAccount;
	}

	////////////////////////
	// Object implementation
	////////////////////////
	
	public String toString() {
		return this.resource + "[" + this.resourceUnitCount + "]";
	}
}
