/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.element.util;

import java.io.InputStream;
import java.io.IOException;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xmloperator.reactive.exception.ReactiveException;

import org.xmloperator.reactive.element.ChildrenConstraint;
import org.xmloperator.reactive.element.ElementBoard;
import org.xmloperator.reactive.element.RootElement;
import org.xmloperator.reactive.element.Element;
import org.xmloperator.reactive.element.Namespace;
import org.xmloperator.reactive.element.ParentElement;
import org.xmloperator.reactive.element.Rebuilder;
import org.xmloperator.reactive.element.RWParentElement;
import org.xmloperator.reactive.element.TranslatableObject;

import org.xmloperator.reactive.element.impl.ParentElementImpl;
import org.xmloperator.reactive.element.impl.StringDataElementImpl;
import org.xmloperator.reactive.element.impl.LongDataElementImpl;
import org.xmloperator.reactive.element.impl.DoubleDataElementImpl;
import org.xmloperator.reactive.element.impl.ObjectDataElementImpl;

import org.xmloperator.reactive.element.constraint.ChildrenConstraintFactory;

/**
 * A DocumentReader reads a RootElement as an XML document.
 */
public final class DocumentReader extends DefaultHandler {
	
	/**
	 * Reads a RootElement as an XML 1.0 document.
	 * @param inputStream an InputStream. It is required.
	 * @param rootElement a RootElement.
	 */
	public static void readDocument(InputStream inputStream,
			RootElement rootElement, ElementBoard elementBoard, long writingKey)
			throws IOException, SAXException {
		if (inputStream == null || rootElement == null
				|| rootElement.getChildCount() != 0)
			throw new IllegalArgumentException();
		SAXParserFactory parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(true);
		SAXParser parser;
		try {
			parser = parserFactory.newSAXParser();
		} catch (ParserConfigurationException exception) {
			exception.printStackTrace();
			throw new ReactiveException("ParserConfigurationException");
		}
		parser.parse(
			inputStream,
			new DocumentReader(
				rootElement,
				elementBoard,
				writingKey));
		inputStream.close();
	}
	
	private final RootElement rootElement;
	private final ElementBoard elementBoard;
	private final long writingKey;
	private ParentElement currentParentElement = null;
	private String datatypeValue = null;
	private StringBuffer sb = null;
	private Rebuilder rebuilder = null;
	
	private DocumentReader(
			RootElement rootElement,
			ElementBoard elementBoard,
			long writingKey) {
		this.rootElement = rootElement;
		this.elementBoard = elementBoard;
		this.writingKey = writingKey;
	}
	
	private RWParentElement rwCurrentParentElement() {
		return (RWParentElement)this.elementBoard.rwElement(
			this.currentParentElement, this.writingKey);
	}
	
	////////////////////////////////
	// ContentHandler implementation
	////////////////////////////////

	public void startPrefixMapping (String prefix, String uri)
			throws SAXException {
		if (!uri.equals(Element.REACTIVE_NAMESPACE))
			this.rootElement.addNamespace(new Namespace(uri, prefix));
	}

	public void startElement(
			String namespaceURI, String localName, String rawName,
			Attributes atts) throws SAXException {
		try {
			if (namespaceURI != null)
				namespaceURI = namespaceURI.intern();
			if (this.currentParentElement == null) {
				if (namespaceURI != this.rootElement.getNamespaceURI()
						|| !localName.equals(this.rootElement.getLocalName()))
					throw new ReactiveException(
						"This is not the expected root element");
				this.currentParentElement = this.rootElement;
				return;
			}
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
		try {
			if (this.rebuilder != null) {
				this.rebuilder.startElement(
					namespaceURI, localName, rawName, atts);
				return;
			}
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
		try {
			this.datatypeValue = atts.getValue("datatype");
			if (this.datatypeValue == null) {
				String ccValue = atts.getValue("cc");
				ChildrenConstraint childrenConstraint = null;
				if (ccValue != null)
					childrenConstraint =
						ChildrenConstraintFactory.childrenConstraint(
							ccValue);
				ParentElement newElement =
					new ParentElementImpl(
						namespaceURI, localName, childrenConstraint);
				this.rwCurrentParentElement().appendChild(newElement);
				this.currentParentElement = newElement;
				return;
			}
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
		try {
			if (this.datatypeValue.equals("object")) {
				this.datatypeValue = null;
				String rebuilderClassName = atts.getValue("class");
				if (rebuilderClassName == null)
					throw new ReactiveException(
						"Object data element without class attribute");
				Class<?> rebuilderClass = Class.forName(rebuilderClassName);
				this.rebuilder =
					(Rebuilder)rebuilderClass.newInstance();
				return;
			}
			this.sb = new StringBuffer();
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
	}

	public void endElement(
			String namespaceURI, String localName, String rawName)
			throws SAXException {
		try {
			if (namespaceURI != null)
				namespaceURI = namespaceURI.intern();
			if (this.rebuilder != null) {
				TranslatableObject objectValue =
					this.rebuilder.endElement(namespaceURI, localName, rawName);
				if (objectValue != null) {
					this.rebuilder = null;
					Element newElement =
						new ObjectDataElementImpl(
							namespaceURI, localName, objectValue);
					this.rwCurrentParentElement().appendChild(newElement);
				}
				return;
			}
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
		try {
			if (this.datatypeValue == null) {
				this.currentParentElement = this.currentParentElement.getParent();
				return;
			}
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
		try {
			if (!this.datatypeValue.equals("object")) {
				String stringValue = this.sb.toString();
				this.sb = null;
				Element newElement;
				if (this.datatypeValue.equals("string"))
					newElement =
						new StringDataElementImpl(
							namespaceURI, localName, stringValue);
				else if (this.datatypeValue.equals("long")) {
					long longValue = Long.parseLong(stringValue);
					newElement =
						new LongDataElementImpl(
							namespaceURI, localName, longValue);
				}
				else if (this.datatypeValue.equals("double")) {
					double doubleValue = Double.parseDouble(stringValue);
					newElement =
						new DoubleDataElementImpl(
							namespaceURI, localName, doubleValue);
				}
				else
					throw new ReactiveException("Unknown datatype value");
				this.rwCurrentParentElement().appendChild(newElement);
			}
			this.datatypeValue = null;
		}
		catch (Exception exception) {
			throw new SAXException(exception);
		}
	}

	public void characters (char ch[], int start, int length)
			throws SAXException {
		if (this.rebuilder != null)
			this.rebuilder.characters(ch, start, length);
		else
			sb.append(ch, start, length);
	}
}
