/*
 * Copyright (c) 2008, 2011 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;

import org.xmloperator.reactive.exception.BoardAlreadyAttachedException;
import org.xmloperator.reactive.exception.NotOwnEventException;
import org.xmloperator.reactive.exception.ReactiveException;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.Board;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.MessageEvent;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.PrintStream;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveEntity;
import org.xmloperator.reactive.ReactiveMachine;

/**
 * ReactiveCommunity implementation.
 */
public final class ReactiveCommunityImpl implements ReactiveCommunity {
	
	/**
	 * Group of entities whose threads are released at the same time.
	 */
	class EntityGroup {
		private final boolean isStop;
		private final AtomicInteger entityCount = new AtomicInteger();
		private final AtomicInteger entityWaitingCount = new AtomicInteger();
		private volatile int entityToWakeUpCount = 0;
		private volatile boolean isGroupReleased = false;
		/**
		 * Constructor
		 * 
		 * @param isStop
		 *            true for a group of stopped entities, false for a group of
		 *            suspended entities.
		 */
		EntityGroup(boolean isStop) {
			this.isStop = isStop;
		}

		/**
		 * Returns true if some entity of this group is yet waking up, false
		 * elsewhere.
		 * 
		 * @return true if some entity of this group is yet waking up, false
		 *         elsewhere.
		 */
		final boolean isGroupWakingUp() {
			if (!this.isGroupReleased)
				return false;
			if (this.entityToWakeUpCount != 0)
				return true;
			this.isGroupReleased = false;
			return false;
		}
		
		/**
		 * One entity of this group achieves its cycle. An automata cycle begins
		 * if no more entity (of any group) is active.
		 */
		final void oneEntityAchieveItsCycle(ReactiveEntity reactiveEntity) {
			/*
			 * Verify group release synchronization.
			 */
			if (this.isGroupReleased)
				throw new ReactiveException(reactiveEntity + " : bad group("
						+ (this.isStop ? "stop" : "suspend")
						+ ") release synchronization");
			/*
			 * Counts the calling entity as belonging to this group.
			 */
			this.entityCount.incrementAndGet();
			/*
			 * The calling entity is no more active.
			 */
			if (decrementActivityCount(reactiveEntity) != 0) {
				/*
				 * Others entities (of any group) are active. The calling entity
				 * must wait.
				 */
				this.waitForRelease(reactiveEntity);
				return;
			}
			/*
			 * The activity of every entity of every group is achieved. It is
			 * time for a new cycle. One has to decide the type of the new
			 * cycle.
			 */
			boolean isNewCycleForSuspendedEntities = theNextCyclewillConcernOnlySuspendedEntities();
			if (this.isStop & isNewCycleForSuspendedEntities) {
				/*
				 * Starts a new cycle for suspended entities. The calling
				 * entity, a stopped one, is not concerned.
				 */
				startNewCycle(true);
				/*
				 * The calling entity must wait. Others entities are active.
				 */
				this.waitForRelease(reactiveEntity);
				return;
			}
			/*
			 * Counts the calling entity as waiting.
			 */
			this.entityWaitingCount.incrementAndGet();
			/*
			 * Starts a new cycle, which ends with releasing this group.
			 */
			startNewCycle(isNewCycleForSuspendedEntities);
			/*
			 * No wait for the calling entity, which is the only active entity
			 * at this time.
			 */
		}

		/**
		 * Waits for release of this group.
		 */
		private final void waitForRelease(ReactiveEntity reactiveEntity) {
			if (this.isGroupReleased)
				throw new ReactiveException(reactiveEntity + " : prematured release.");
			synchronized (this) {
				this.entityWaitingCount.incrementAndGet();
				this.entityToWakeUpCount++;
				while (!this.isGroupReleased) {
					try {
						this.wait();
						/*
						 * Ends by notifyAll in function release().
						 */
					} catch (InterruptedException exception) {}
				}
				/*
				 * Counts the calling entity as waked up.
				 */
				this.entityToWakeUpCount--;
			}
		}
		
		/**
		 * Returns true if all entities of the group are waiting, false elsewhere.
		 * @return true if all entities of the group are waiting, false elsewhere.
		 */
		private boolean areAllEntitiesWaiting() {
			synchronized (this) {
				return this.entityWaitingCount.get() == this.entityCount.get();
			}
		}

		/**
		 * Resets the group.
		 */
		private void reset() {
			this.entityCount.set(0);
			this.entityWaitingCount.set(0);
		}
		
		/**
		 * Releases all entities of this group for a new cycle.
		 */
		void release() {
			this.isGroupReleased = true;
			synchronized (this) {
				this.notifyAll();
			}
		}
	}

	//////////////
	// Static data
	//////////////

	private static AtomicInteger reactiveCommunityCount = new AtomicInteger();
	private static PrintStream trace = null;
	private static boolean isFullTrace = false;

	/////////////////
	// Static methods
	/////////////////

	public static void setTrace(PrintStream trace, boolean isFullTrace) {
		ReactiveCommunityImpl.trace = trace;
		ReactiveCommunityImpl.isFullTrace = isFullTrace;
	}

	/////////////////////////////
	// Shared final instance data
	/////////////////////////////

	final String communityId = ("C" + String
			.valueOf(ReactiveCommunityImpl.reactiveCommunityCount
					.incrementAndGet())).intern();
	final MessageEventImpl actionMessageEvent;

	////////////////////////
	// Private instance data
	////////////////////////

	/*
	 * Cycle.
	 */
	private volatile int status = STATUS_IDLE;
	private volatile long instant = 0;
	private volatile long cycle = 0;
	/*
	 * Synchronization.
	 */
	private final Object synchroStatus = new Object();
	/*
	 * Entities.
	 */
	private final List<ReactiveEntity> entityToConnectList = new ArrayList<ReactiveEntity>();
	private final Set<ReactiveEntity> connectedEntitySet = new HashSet<ReactiveEntity>();
	private final EntityGroup suspendedEntityGroup = new EntityGroup(false);
	private final EntityGroup stoppedEntityGroup = new EntityGroup(true);
	/*
	 * During agent activity (i.e. automata waiting) we have : activeEntityCount
	 * + suspendedEntityCount + stoppedEntityCount == count of connected agent.
	 */
	private final AtomicInteger activeEntityCount = new AtomicInteger();
	/*
	 * Events.
	 */
	private final List<Event> generatedEventList = new ArrayList<Event>();
	private final List<MessageEvent> generatedMessageEventList = new ArrayList<MessageEvent>();
	private final List<Event> broadcastedEventList = new ArrayList<Event>();
	private final List<Event> unmodifiedBroadcastedEventList = Collections
			.unmodifiableList(this.broadcastedEventList);
	/*
	 * Board
	 */
	private volatile Board futureBoard = null;
	private volatile Board board = null;
	private volatile long boardPermissionKey = 0;
	/*
	 * ReactiveMachine.
	 */
	private final ReactiveMachine reactiveMachine;
	
	/**
	 * Constructor.
	 */
	public ReactiveCommunityImpl(ReactiveMachine reactiveMachine) {
		if (reactiveMachine == null)
			throw new IllegalArgumentException();
		this.reactiveMachine = reactiveMachine;
		this.actionMessageEvent = new MessageEventImpl(this);
	}

	/**
	 * Returns true if the next cycle will occur in the same instant because
	 * there are suspended entities and instantaneous events to broadcast.
	 * Return false if the next cycle will concern a new instant.
	 * 
	 * @return true if the next cycle will occur in the same instant because
	 *         there are suspended entities and instantaneous events to
	 *         broadcast, false elsewhere.
	 */
	private boolean theNextCyclewillConcernOnlySuspendedEntities() {
		return (this.suspendedEntityGroup.entityCount.get() != 0)
				&& !this.generatedEventList.isEmpty();
	}

	/**
	 * Starts a new cycle automata. It will be a cycle for suspended entities if
	 * events and suspended entities exist.
	 */
	private void startNewCycle() {
		this.startNewCycle(this.theNextCyclewillConcernOnlySuspendedEntities());
	}

	/**
	 * Starts a new automata cycle.
	 * 
	 * @param isNewCycleForSuspendedEntities
	 *            if true then starts a cycle for suspended entities else starts
	 *            a cycle for all entities.
	 */
	private void startNewCycle(boolean isNewCycleForSuspendedEntities) {
		/*
		 * Status: ACTIVE -> SUSPENDED
		 */
		if (this.status != STATUS_ACTIVE)
			throw new ReactiveException("Automata status should be ACTIVE.");
		this.status = STATUS_SUSPENDED;
		/*
		 * Verify no more entity activity.
		 */
		if (this.activeEntityCount.get() != 0)
			throw new ReactiveException("There should be no active entity.");
		/*
		 * Verify : connectedEntities = suspendedEntities + stoppedEntities.
		 */
		if (this.suspendedEntityGroup.entityCount.get()
				+ this.stoppedEntityGroup.entityCount.get() != this.connectedEntitySet
				.size())
			throw new ReactiveException(
					"Connected entities should be suspended or stopped.");
		/*
		 * New Cycle.
		 */
		this.cycle++;
		/*
		 * Is it a cycle for suspended entities ?
		 */
		if (isNewCycleForSuspendedEntities) {
			/*
			 * Broadcasts events.
			 */
			this.broadcastedEventList.clear();
			this.broadcastedEventList.addAll(this.generatedEventList);
			this.generatedEventList.clear();
			/*
			 * Waits for suspended entities are waiting.
			 */
			while (!this.suspendedEntityGroup.areAllEntitiesWaiting())
				Thread.yield();
			/*
			 * Suspended entities become active again.
			 */
			this.activeEntityCount.set(this.suspendedEntityGroup.entityCount
					.get());
			/*
			 * Reset the group of suspended entities.
			 */
			this.suspendedEntityGroup.reset();
			/*
			 * Status: SUSPENDED -> ACTIVE
			 */
			this.status = STATUS_ACTIVE;
			/*
			 * Release suspended entities.
			 */
			this.suspendedEntityGroup.release();
			return;
		}
		/*
		 * For this instant, potentials events are lost, since there is no
		 * destination entity.
		 */
		this.broadcastedEventList.clear();
		this.generatedEventList.clear();
		/*
		 * Status: SUSPEND -> INSTANT_TRANSITION -> BEFORE_ACTIVE
		 */
		this.status = STATUS_INSTANT_TRANSITION;
		this.instant += 1;
		if (trace != null)
			trace.nextInstant(this);
		this.status = STATUS_BEFORE_ACTIVE;
		/*
		 * Connects requesting entities.
		 */
		while (!this.entityToConnectList.isEmpty()) {
			ReactiveEntity reactiveEntity;
			synchronized (this.entityToConnectList) {
				reactiveEntity = (ReactiveEntity) this.entityToConnectList.remove(0);
			}
			this.connectedEntitySet.add(reactiveEntity);
		}
		/*
		 * Connected entities become active.
		 */
		this.activeEntityCount.set(this.connectedEntitySet.size());
		/*
		 * Processes action message.
		 */
		this.board = this.futureBoard;
		this.processActionMessages();
		/*
		 * Is there connected entities ?
		 */
		if (this.activeEntityCount.get() == 0) {
			/*
			 * There is no connected entity. Eventual messages are lost.
			 */
			for (MessageEvent messageEvent : this.generatedMessageEventList)
				((MessageEventImpl)messageEvent).startNewInstant(false);
			this.generatedMessageEventList.clear();
			/*
			 * Status: BEFORE_ACTIVE -> IDLE
			 */
			this.status = STATUS_IDLE;
			if (trace != null)
				trace.nextInstant(null);
		} else {
			/*
			 * There are connected entities. Messages (if any) are moved in order to
			 * be broadcasted during the first suspend cycle of this instant.
			 */
			this.generatedEventList.addAll(this.generatedMessageEventList);
			for (MessageEvent messageEvent : this.generatedMessageEventList)
				((MessageEventImpl)messageEvent).startNewInstant(true);
			this.generatedMessageEventList.clear();
			/*
			 * Waits for suspended and stopped entities are waiting.
			 */
			while (!this.suspendedEntityGroup.areAllEntitiesWaiting())
				Thread.yield();
			while (!this.stoppedEntityGroup.areAllEntitiesWaiting())
				Thread.yield();
			/*
			 * Reset all groups of entities.
			 */
			boolean existsSuspendedEntities = this.suspendedEntityGroup.entityCount
					.get() != 0;
			boolean existsStoppedEntities = this.stoppedEntityGroup.entityCount
					.get() != 0;
			this.suspendedEntityGroup.reset();
			this.stoppedEntityGroup.reset();
			/*
			 * Status: BEFORE_ACTIVE -> ACTIVE
			 */
			this.status = STATUS_ACTIVE;
			/*
			 * Releases all entities. No Event is return.
			 */
			if (existsSuspendedEntities)
				this.suspendedEntityGroup.release();
			if (existsStoppedEntities)
				this.stoppedEntityGroup.release();
		}
	}

	/**
	 * Decrements the active entity count.
	 * 
	 * @return the new count.
	 */
	private int decrementActivityCount(ReactiveEntity reactiveEntity) {
		int activityCount = this.activeEntityCount.decrementAndGet();
		if (activityCount < 0)
			throw new ReactiveException("Active entity count cannot be negative.");
		return activityCount;
	}

	private void processActionMessages() {
		List<Message> messageList = this.actionMessageEvent
				.getGeneratedMessageList();
		if (messageList.isEmpty())
			return;
		for (int priority = Action.PRIORITY_HIGH; priority >= Action.PRIORITY_LOW; priority--)
			for (Message message : messageList)
				if (message.isActionMessage()) {
					ActionMessageImpl actionMessage = (ActionMessageImpl) message;
					Action action = actionMessage.getAction();
					if (action.getPriority() == priority
							&& !action.isExecuted()) {
						try {
							action.execute(this.boardPermissionKey,
									this.instant);
						} catch (ReactiveException exception) {
							actionMessage.setException(exception);
						}
					}
				}
	}

	private void waitForAllGroupsReadyForANewCycle() {
		while (this.suspendedEntityGroup.isGroupWakingUp())
			Thread.yield();
		while (this.stoppedEntityGroup.isGroupWakingUp())
			Thread.yield();
	}
	
	////////////////////////////////////////////
	// Methods to be called by ReactiveEntityImpl
	////////////////////////////////////////////

	/**
	 * Connects a given ReactiveEntity.
	 * @param reactiveEntity an entity to connect to this community.
	 */
	void connect(ReactiveEntity reactiveEntity) {
		synchronized (this.entityToConnectList) {
			this.entityToConnectList.add(reactiveEntity);
		}
		synchronized (this.synchroStatus) {
			if (this.status == STATUS_IDLE) {
				this.status = STATUS_BEFORE_ACTIVE;
				this.entityToConnectList.remove(reactiveEntity);
				this.connectedEntitySet.add(reactiveEntity);
				this.activeEntityCount.incrementAndGet();
				if (trace != null)
					trace.nextInstant(this);
				this.status = STATUS_ACTIVE;
				return;
			}
		}
		while (!this.connectedEntitySet.contains(reactiveEntity));
			Thread.yield();
		while (this.getStatus() != ReactiveCommunity.STATUS_ACTIVE)
			Thread.yield();
	}

	/**
	 * Generates an Event to this community.
	 * @param reactiveEntity the sender reactiveEntity.
	 * @param event an Event.
	 */
	void generateEvent(ReactiveEntity reactiveEntity, Event event) {
		if (event.isMessageEvent())
			throw new ReactiveException("This Event is not a base Event.");
		if (event.getCommunity() != this)
			throw new NotOwnEventException();
		if (((EventImpl) event).generate(this.instant)) {
			synchronized (this.generatedEventList) {
				this.generatedEventList.add(event);
			}
			if (trace != null)
				trace.println(this + "." + this.instant + " " + reactiveEntity
						+ " generate " + event);
		}
	}

	/**
	 * Generates a MessageEvent to this community.
	 * @param reactiveEntity the sender ReactiveEntity.
	 * @param messageEvent the MessageEvent to generate.
	 * @param message a Message.
	 */
	void generateMessage(ReactiveEntity reactiveEntity,
			MessageEvent messageEvent, Message message) {
		if (messageEvent.getCommunity() != this)
			throw new NotOwnEventException();
		MessageEventImpl messageEventI = (MessageEventImpl) messageEvent;
		if (messageEventI.generate(this.instant))
			synchronized (this.generatedMessageEventList) {
				this.generatedMessageEventList.add(messageEvent);
			}
		if (message != null)
			messageEventI.addGeneratedMessage(message);
		if (trace != null && (message == null || message.isTraceable()))
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " generate " + messageEvent + " " + message);
	}

	List<Event> suspend(ReactiveEntity reactiveEntity) {
		if (trace != null && isFullTrace)
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " suspend");
		this.waitForAllGroupsReadyForANewCycle();
		this.suspendedEntityGroup.oneEntityAchieveItsCycle(reactiveEntity);
		return this.unmodifiedBroadcastedEventList;
	}

	void stop(ReactiveEntity reactiveEntity) {
		if (trace != null && isFullTrace)
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " stop");
		this.waitForAllGroupsReadyForANewCycle();
		this.stoppedEntityGroup.oneEntityAchieveItsCycle(reactiveEntity);
	}

	void disconnect(ReactiveEntity reactiveEntity) {
		synchronized (this.connectedEntitySet) {
			this.connectedEntitySet.remove(reactiveEntity);
		}
		if (trace != null && isFullTrace)
			trace.println(this + "." + this.instant + " " + reactiveEntity
					+ " disconnected");
		this.waitForAllGroupsReadyForANewCycle();
		if (this.decrementActivityCount(reactiveEntity) == 0)
			this.startNewCycle();
	}

	// /////////////////////////////////
	// ReactiveCommunity implementation
	// /////////////////////////////////

	public int getStatus() {
		return this.status;
	}

	public long getInstant() {
		return this.instant;
	}

	public Event createEvent() {
		return new EventImpl(this);
	}

	public MessageEvent createMessageEvent() {
		return new MessageEventImpl(this);
	}

	public MessageEvent getActionMessageEvent() {
		return this.actionMessageEvent;
	}

	public Board getBoard() {
		return this.board;
	}

	public void setBoard(Board board, long permissionKey) {
		if (this.futureBoard != null)
			throw new BoardAlreadyAttachedException();
		this.futureBoard = board;
		this.boardPermissionKey = permissionKey;
	}

	public String launchAgent(Agent agent) {
		return this.reactiveMachine.launchAgent(agent, this);
	}

	// //////////////////////
	// Object implementation
	// //////////////////////

	public String toString() {
		return this.communityId;
	}
}
